/* RoboR Control program   July 26 2001

 This program is compiled using the Linux shareware C2C compiler 
 available at: http://www.geocities.com/SiliconValley/Network/3656/
 
 The output .asm code is assembled with the Linux freeware assembler
 gpasm avaiable at:  http://gpasm.sourceforge.net 
 
         
          
V 23
V 20
-----------------------------------------------------------          
          
 
 */

/* Be sure to set the configuration word to 3f72 when programming the chip
   It's done in the makefile */



#define FOSC   16000000
#pragma  CLOCK_FREQ  16000000

#include "16f876.h"

//#define TMR0DIV   256-(FOSC/(8000*16)) //Compiler will not do this math!
//Divider constant for 8,000 interrupts per second

//For 8000 interrupts/sec
#define TMR0DIV 133

  

#define BAUD 9600
//#define BAUDDIV (FOSC/(BAUD*64)) - 1
#define BAUDDIV 26



/* these values written into ADCCON0 
   set the A-D channel mux address 
   and converison clock to 1/32 osc.
   This makes a 32uS converson time*/
#define  AD_0           0x81    /* ch 0 left front - right rear edge sensor*/
#define  AD_1           0x89    /* ch 1 right front - left rear edge sensor*/
#define  AD_LMFB        0x99    /* channel 3 Left motor feedback*/
#define  AD_RMFB        0x91    /* channel 2 Right motor feedback*/
#define  AD_4           0xa1     /* ch 4  Drive motor current 1v/amp*/
#define  AD_5           0xa9     /* ch 5      */
#define  AD_6           0xb1     /* ch 6 Battery voltage divided by 10*/
#define  AD_7           0xb9     /* ch 7      */    

char tx_buf[34]@0x24;     /* Tx serial port buffer */
char iavg[4]@0x20;


char spi_ptr@0x46;         /* SPI buffer pointer */
char spi_buf[15]@0x47 ;     /* SPI port buffer for sonar data */
//spi_buf offsets...

#define t_echo_LF 3
#define a_echo_LF 4
#define t_echo_RF 5
#define a_echo_RF 6
#define t_echo_LR 7
#define a_echo_LR 8
#define t_echo_RR 9
#define a_echo_RR 10
#define pot_value 11
#define tilt_dir_val 12
#define battery_val 13 

#define raw_IR 255

#define TILT 6    /*Bit 6 is low true tilt status*/
#define DIR 7     /*Bit 7 is direction status Forward=LOW */

/* Analog inputs */
char  speed;

char  rmfb;
char  lmfb;

//char  mode;


char  right_speed,right_prm;
char  left_speed,left_prm;

char  bot_mode;
char  bump_sw;
char  IR_sens;
 

#define MODE_STOP 0

#define MODE_RUN 1

#define MODE_GOHOME 7
#define MODE_NULL 0xff

#define REV0 0xf9
#define REV1 0xf6 
#define REV2 0xf2
#define REV3 0xed
#define REV4 0xe7
#define REV5 0xd8
#define FWD0 7
#define FWD1 10
#define FWD2 14
#define FWD3 19
#define FWD4 25
#define FWD5 40




char  prescale;
char  timer;
char  prm_ctl;
/* bits in prm_ctl */
#define  R_fwd 0
#define  L_fwd 1


//Port C bits
//#define TILT 0       /* input, Low true when tilted */
#define LED  1       /* Low true output turns on LED */
#define IR_ENABLE 2  /* Low true output turns on IR sensor LEDs */


char  workB;

/* Port B motor control bits */

#define  LeftEnable 5    /* Left Motor pwm (L293D enable enable)*/
#define  RightEnable 4   /* Right Motor pwm (L293D enable enable)*/
#define  FwdLeft  3     /*Left Motor forward */
#define  RevLeft  2     /*Left Motor reverse*/
#define  FwdRight 1     /*Right Motor forward */
#define  RevRight 0     /*Right Motor reverse */



#define  P_width  22    /*Motor pulse width in interrupt ticks*/
char tx_ptr,rx_ptr,txstate,rxstate;
char  crcH,crcL;

char r_speed;
char l_speed;
char l_integrate;
char r_integrate;
char dir_ctl;

char testmode;
char speed_to_pulse_rate(char speed);


char timer2,startup_timer,motor_timer,tilt_dir, escape_timer;
char tilt_memory, border_memory, pushback_memory, cruse_memory; 

char  counter1;
char   runTimer_H;  /* measures run time in seconds */
char   runTimer_L;
char spi_stat;
char  battery;
char  drive_current, drive_current_avg;


char flags, arb_code;
char debug;
char mon_dir_ctrl;
char  lf_rr_sens, rf_lr_sens;

char cruse_output, chase_output , current_limit_output , border_output, tilt_output;
char check_battery_output, continue_output, arbitrate_output, pushed_back_output;



char lf_range,lf_ampl, rf_range, rf_ampl, lr_range, lr_ampl, rr_range, rr_ampl, pot_val;

char lf_avg,rf_avg,lr_avg,rr_avg, f_ratio, r_ratio;

char divide10;
char divide100;
char lr_ratio;




//Bits in flags variable
#define tick100 0
#define direction 1
#define r_odo_last 2
#define stalled 3
#define spin_left 4
#define spin_right 5
#define attack 6


char rnd;            /*random number*/
char ctr;           /* a counter */
char temp;




//----------------------------------------------------------




#define COUNT 1      /* offset to count field in tx/rx buffers */
#define ADDRESS 0    /* offset to address field in tx/rx buffers */
#define DATA 2       /* offset to start of data in tx/rx buffers */





char test_hi@0xfe;

char detect_line();
char neg(char x);
char integrate(char speed,char mfb,char integ);






//------------------------------------------------------------------

/* Interrupts from Timer 0 occur at about 8,000 per second. */
void interrupt(void)
{
   char R_pulse,L_pulse,motor_avg;
   char i,l,r,spidata;
   
   
   if(PIR1 & 8){        //if SPI port interrupt...
       
       set_bit(STATUS,RP0);     //Bank1
       if(SSPSTAT & 1){       //Check BF buffer full status bit
          clear_bit(STATUS,RP0); //bank 0
          spidata = SSPBUF;
          if(spi_ptr != 0) spi_buf[spi_ptr++] = spidata ;
          if(spi_ptr == 0){
             if(SSPBUF == 0x7e)    //Be sure 0x7e is first byte
               spi_buf[spi_ptr++] = spidata ;
          }

          
          
       }
         clear_bit(STATUS,RP0); //bank 0
         clear_bit(PIR1,SSPIF);  //Clear this SPI interrupt
         if(spi_ptr > 13){        //Got all 14 bytes?
             
             spi_ptr = 0;
             spi_stat = 0xff;
                      
              lf_range = spi_buf[t_echo_LF + 0];
              lf_ampl  = spi_buf[a_echo_LF + 0];

              rf_range = spi_buf[t_echo_RF + 0];
              rf_ampl  = spi_buf[a_echo_RF + 0];

              lr_range = spi_buf[t_echo_LR + 0];
              lr_ampl  = spi_buf[a_echo_LR + 0];

              rr_range = spi_buf[t_echo_RR + 0];
              rr_ampl  = spi_buf[a_echo_RR + 0];

              pot_val  = spi_buf[pot_value + 0];
              battery  = spi_buf[battery_val + 0];
              tilt_dir = spi_buf[tilt_dir_val + 0];
         
         
              
         
         }
         
         
   }

   if (INTCON & 4) {   //Check for a TMR0 interrupt

      

      clear_bit(INTCON,2);      //Clear timer 0 interrupt flag
      TMR0 = TMR0DIV + TMR0;       //restart timer 0
      

      if(prescale-- == 0){
         
         prescale = 79;
         if(timer != 0) timer--;    //10 ms pertick timer
         if(timer2 != 0) timer2--;
         if(motor_timer != 0) motor_timer--;
         
      
                 
         if(divide10++ >= 9){
             
             if(startup_timer != 0)    startup_timer--; //100ms per tick timers
             if(escape_timer != 0)     escape_timer--;
             if(cruse_memory != 0)     cruse_memory--;
             if(pushback_memory != 0)  pushback_memory--;
             if(border_memory != 0 )   border_memory--;
             if(tilt_memory != 0 )     tilt_memory--;
             
             divide10 = 0;
             set_bit(flags,tick100);
                                       
         }

         if(divide100++ >= 99){  // 1 second per tick timer
            runTimer_L++;
            if(runTimer_L == 0) runTimer_H++;
            divide100 = 0;

            
         }

         

      }

      PORTB = workB;       //output the prm motor control bits to port B
      
      workB = 0;
      

         /* adjust right pulse rate modulator */
         if (right_prm < P_width) {
            set_bit(workB,RightEnable);
            if (prm_ctl & (1 << R_fwd))
               set_bit(workB,FwdRight);
            else
               set_bit(workB,RevRight);
         }
         if (right_prm-- == 0) {
            right_prm = speed_to_pulse_rate(right_speed) + P_width;
            if (right_speed > 127) clear_bit(prm_ctl,R_fwd);
            else set_bit(prm_ctl,R_fwd);
         }

         /*adjust left pulse rate modulator */
         if (left_prm < P_width) {
            set_bit(workB,LeftEnable);
            if (prm_ctl & (1 << L_fwd))
               set_bit(workB,FwdLeft);
            else
               set_bit(workB,RevLeft);
         }

         if (left_prm-- == 0) {
            left_prm = speed_to_pulse_rate(left_speed) + P_width;
            if (left_speed > 127) clear_bit(prm_ctl,L_fwd);
            else set_bit(prm_ctl,L_fwd);
         }


      }
   
    if (bot_mode == MODE_STOP) workB = 0;    //Stop the bot.
      
    if ((l_speed == 0) && (r_speed == 0)) workB = 0;
    if(testmode == 1) workB = 0;
      
      //Make a new random number
      if (rnd & 0x80) { //rnd msb is 1
            asm{
               movlw 1ch ;
               xorwf _rnd,F;
               bsf   STATUS,C;
               rlf   _rnd; //8 bit left shift
              
            }
         } else {    //rnd msb  is 0
            asm{
               bcf   STATUS,C;
               rlf   _rnd; //8 bit left shift
               
            }
         }
       
      

            

   }
//--------------------------------------------------------------
short mpy(char a, char b)
{
   return a * b;
}
//---------------------------------------------------------------
char divide(char a, char b)
{
   return a / b;
}

   //-----------------------------------------------------------------------

   // compute 16 bit CRC on 1 data byte. 
   //   crcH and crcL  are global variables which accumulate the 16 bit CRC. 
   //   They must be cleared before starting to compute 16 bit CRC on a string.
   //   The final output value will be in crcH and crcL.
   
   void crc(char data){

      char i;
     // short crc16;

      for (i=0;i<8;i++) {                //Do all 8 bits in the data byte
         if (data & 0x80) crcH ^= 0x80; //if data bit is 1 then flip crc msb
         data = data << 1;             // get next data bit
         if (crcH & 0x80) { //crc msb is 1
            asm{
               bcf   STATUS,C;
               rlf   _crcL; //16 bit left shift
               rlf   _crcH;
               movlw 80h;
               xorwf _crcH,F; //xor with 0x8005
               movlw 5;
               xorwf _crcL,F;

            }
         } else {    //crc msb  is 0
            asm{
               bcf   STATUS,C;
               rlf   _crcL; //16 bit left shift
               rlf   _crcH;
            }
         }
      }

   }
   //-----------------------------------------------------------------------
   void set_tx_buffer_crc(){
      int j = tx_buf[COUNT + 0] + 2; //get data size
      int i;

      crcH = 0;         //Clear crc16
      crcL = 0;
      for (i=0;i<j;i++) crc(tx_buf[i]); //Compute crc16 over buffer
      tx_buf[i++] = crcH;      //tack the crc bytes on the end of the buffer
      tx_buf[i++] = crcL;
      txstate = 0;            //State 0 = stopped
   }

   //-----------------------------------------------------------------------
   
   void make_info_pkt(){
      char i,j;
      i = ADDRESS;
      tx_buf[i] = 1;             // send to master
      i = COUNT;
      tx_buf[i] = 30;            // n bytes
      i = DATA;
      tx_buf[i++] = 0x01;        //data type = 1
      tx_buf[i++] = l_speed;  //1
      tx_buf[i++] = r_speed; //2
      tx_buf[i++] = flags;  //3
      tx_buf[i++] = runTimer_H;  //4
      tx_buf[i++] = runTimer_L;  //5

      tx_buf[i++] = f_ratio; //6
      tx_buf[i++] = r_ratio;          //7
      tx_buf[i++] = tilt_dir;   //8 
      tx_buf[i++] = bot_mode;      //9
      
      tx_buf[i++] = battery;       //10 Battery voltage/3
      tx_buf[i++] = mon_dir_ctrl;          //11
      tx_buf[i++] = timer;         //12
      tx_buf[i++] = lf_rr_sens ;   //13  Left front minus right rear line sensor
      tx_buf[i++] = rf_lr_sens ;   //14  Right front minus left rear line sensor
      tx_buf[i++] = drive_current; //15
      tx_buf[i++] = arb_code;      //16
      
      for(j=0;j<14;j++){
         tx_buf[i++] = spi_buf[j];  //16..28 are SPI port buffer data
      }
      
      

      crcH = 0;
      crcL = 0;
      set_tx_buffer_crc();  //compute and add CRC16 bits to end of pkt

      txstate = 1;          //Tell scheduler it's ready to transmit
   }
   
   //----------------------------------------------------------------------




  //send character if tx ready and return true 
  //  else return false if tx not ready 
   char tx_byte(char c){
      if (PIR1 & (1 << TXIF)) {
         TXREG = c; 
         return 1;     
      } else
         return 0;
   }

 
  //-----------------------------------------------------------------------


   void send_tx_buffer(){
      static char j;

      
      switch (txstate) {
      case 1:
      case 2:
      case 3:  if (tx_byte(0xff)) txstate++;
         break;
      case 4:  if (tx_byte(0xf0)) txstate++;
         break;
      case 5:  if (tx_byte(0x55)) txstate++;
         break;
      case 6:  tx_ptr = 0;
         j = tx_buf[COUNT + 0] + 4; //Must use COUNT+0 to force indirect addressing
         txstate++;
         break;

      case 7:  if (tx_byte(tx_buf[tx_ptr])) {
            j--;
            tx_ptr++ ;
            if (j == 0) txstate = 8;
         }
         break;

      case 8:  if (tx_byte(0x33)) {
            txstate = 0;
         }



      }

      
   }

   

      
   //-----------------------------------------------------------------------
   /* Convert a signed value to an absolute value between 127 and 0 
        0 becomes 127
      127 becomes 0
     -128 becomes 0
     
   */
   char speed_to_pulse_rate(char speed)
   {
      if (speed > 127) {      //See if speed is negative !
         return speed & 0x7f;
      } else {
         return speed ^ 0x7f;
      }



   }

   //----------------------------------------------------------------------
   char ad_convert(char channel)
   {  char x;
      
      
      ADCON0 = channel;
      for (x=16;x>0;x--);         //wait at least 20 us for data to stabilize
      
      set_bit(ADCON0,GO_DONE);   //start A-D conversion
      
      while (ADCON0 & (1 << NOT_DONE));  //wait for conversion to complete
      
      return ADRESH+128;          //Return signed result
   }

   //---------------------------------------------------------------------

   void get_AD_values(){
    
    
    
    clear_bit(PORTC,IR_ENABLE) ;             //White line sensor IR emitters ON
    drive_current = ad_convert(AD_4) + 128;  //unsigned drive motor current 1v/amp
    lf_rr_sens = ad_convert(AD_0) + 128;     //left fron minus right rear sensor
    rf_lr_sens = ad_convert(AD_1) + 128;     //right front minus left rear sensor
    set_bit(PORTC,IR_ENABLE);                //Line sensor IR emitters OFF

    

   }

   //----------------------------------------------------------------------
   
  
   void adjust_motor_speed(){
      char i;
    
      rmfb = ad_convert(AD_RMFB); //get right motor feedback
      lmfb = ad_convert(AD_LMFB); //get left motor feedback
    
      for (i=0;i<10;i++) {  //Adjust motor speeds based on feedback values
     
        r_integrate = integrate(r_speed,rmfb,r_integrate);
        l_integrate  = integrate(l_speed,lmfb,l_integrate);
        right_speed = r_integrate;
        left_speed  = l_integrate;
       
      }
   }
  //----------------------------------------------------------------------
 
   //dir_control constants.
   
   #define FWD 1
   #define FWD_LEFT 2
   #define FWD_RIGHT 3
   #define FWD_HARD_LEFT 4
   #define FWD_HARD_RIGHT 5
   #define FWD_MAX 6
   
   #define SPIN_LEFT 10
   #define SPIN_RIGHT 11
   #define STOP 12

   #define REV 16
   #define REV_LEFT 17
   #define REV_RIGHT 18
   #define REV_HARD_LEFT 19
   #define REV_HARD_RIGHT 20
   #define REV_MAX 21
   
   
   void dir_control(char value)
   {
      mon_dir_ctrl = value;
      if(value <= FWD_MAX)  set_bit(flags,direction) ; //Set forward flag
      if(value >= REV) clear_bit(flags,direction);     //or set the reverse flag
      

       switch(value){

       
       case FWD:     l_speed = FWD4;
                     r_speed = FWD4;
                     break;

       case FWD_MAX: l_speed = FWD5;
                     r_speed = FWD5;
                     break;

       case FWD_LEFT:
                     l_speed = FWD2;
                     r_speed = FWD5;
                     break;

       case FWD_RIGHT:
                     l_speed = FWD5;
                     r_speed = FWD2;
                     break;

       case FWD_HARD_LEFT:
                     l_speed = 0;
                     r_speed = FWD5;
                     break;

       case FWD_HARD_RIGHT:
                     l_speed = FWD5;
                     r_speed = 0;
                     break;

       case REV:     l_speed = REV4;
                     r_speed = REV4;
                     break;

       case REV_MAX: l_speed = REV5;
                     r_speed = REV5;
                     break;


       case REV_LEFT:
                     l_speed = REV2;
                     r_speed = REV5;
                     break;

       case REV_RIGHT:
                     l_speed = REV5;
                     r_speed = REV2;
                     break;

       case REV_HARD_LEFT:
                     l_speed = 0;
                     r_speed = REV5;
                     break;

       case REV_HARD_LEFT:
                     l_speed = REV5;
                     r_speed = 0;
                     break;

       case SPIN_LEFT:
                     l_speed = REV4;
                     r_speed = FWD4;
                     break;

       case SPIN_RIGHT:
                     l_speed = FWD4;
                     r_speed = REV4;
                     break;

       case STOP:
      
                     l_speed = 0;
                     r_speed = 0;
                     break;

       }
   }


   
  //---------------------------------------------------------------------------------


   /* Saturating addition.  Result is clipped at +127 or -128 
      instead of overflowing.
      Note the parameters a and b and signed char values (-128..+127)
      but the compiler thinks they are unsigned values (0..255) so
      I had to do some strange looking things below.
      */
  
   char addsat(char a, char b)
   {  char r;
      if (a > 127) goto a_minus;  //See if a is minus
      if (b > 127) goto a_b_diff; //a is plus, check b for minusness

      r = a + b;                //Both a and b are positive numbers
      if (r > 127) r = 127; //clip at +limit
      return r;

      a_minus:
      if (b < 0x80) goto a_b_diff;  //A is minus, now check b and see if it's minus too
      r = a + b;
      if (r < 0x80) r = 0x80;  //clip at -128
      return r; 

      a_b_diff:
      return a + b;     //No clipping needed


   }
   
     //---------------------------------------------------------------------
   /* return the absolute value of x */
   
   char abs(char x){
      if(x > 127){
         x ^= 0xff;
         x++;
      }
      return x;
   }
   
   //-----------------------------------------------------------------------
   char abs_diff(char a, char b)
   {
      
      if(a == 0) return b;
      if(b == 0) return a;
      if(a > b) return (a - b);
         else return ( b - a) ;
      
   }

   //-----------------------------------------------------------------------
   /* return the negative of x */
   
   char neg(char x){
      x ^= 0xff;
      x++;
      return x;
   }
   

   //----------------------------------------------------------------------
   /* return x if its 0 or less else return 0 */
   
   char clip_pos(char x){
      if(x < 128) x = 0;
      return x;
   }
   

   //----------------------------------------------------------------------
   /* return x if it's 0 or higher else return 0 */
   
   char clip_neg(char x){
      if(x > 127) x = 0;
      return x;
   }
   
   //---------------------------------------------------------------------
   //return the a 16 bit value
   char sign16(int x){
      char rv;
      rv = 0;
      if(x & 0x8000) rv = 1;
      return rv;
   }
   //---------------------------------------------------------------------

   //Limit x to +/- l
   char limit(char x, char l){

      if(x & 0x80){
         if( x < neg(l)) return neg(l);
         else return x;
      }
      if( x > l) return l;
      else return x;


   }
   //-----------------------------------------------------------

   char divideby2(char x){
      if(x < 128) return (x >> 1);
      else{
         return (x >> 1) | 0x80;
      }
   }

   //-------------------------------------------------------------------
   char ratio(char a, char b)
   {  short d;
      char c;
      if((a == 0) && (b == 0)) return 0;
      c = abs_diff(a,b);
      d = c * 256;
      d = d -1;
      if(a > b) c = a ; else c = b;
      if(c == 0) return 255;
       else return (d / c);
   }

   //----------------------------------------------------------------------
   char integrate(char speed,char mfb,char integ)
   {
      speed = speed + 128;   //convert to unsigned
      mfb = mfb + 128;
      if (speed > mfb) {
         integ++;
         if (integ == 0x80) integ--;  //Clip at +127
         return integ;
      } else {
         integ--;
         if (integ == 0x7f) integ++;  //Clip at -128
         return integ;
      }



   }
   //-----------------------------------------------------------------------
   #define IR_MAX 164
   #define IR_MIN 89  
   /*Reverse directon when we see the white border line */
   void check_border()
   {  char action;
      static char lock;
      
      action = 0;

      if(flags & (1 << direction)){  //going forward?
         if(lock != 1){
            if(lf_rr_sens > IR_MAX)
               action = REV_LEFT;
         

            if(rf_lr_sens > IR_MAX)
               action = REV_RIGHT;
         

            if((lf_rr_sens >  IR_MAX) && (rf_lr_sens > IR_MAX))
               action = REV;     //Put it in reverse

            if(action) lock = 1;
         }


         
      }else    //must be going in reverse
         {
         if(lock != 2){
            if(lf_rr_sens < IR_MIN)
               action = FWD_RIGHT;
         

            if(rf_lr_sens < IR_MIN)
               action = FWD_LEFT;
         
         
            if((lf_rr_sens <  IR_MIN) && (rf_lr_sens < IR_MIN))
               action = FWD;   //go forward

            if(action) lock = 2;
         }
      }


     if(border_memory) return;

     if(action){
        border_memory = 5;      //500ms timeout
        border_output = action;
     }

     if(border_memory == 0){
         border_output = 0;
         lock = 0;
     }


   }
 //------------------------------------------------------------------------
  
   void cruse()
   {  
      cruse_output = FWD;
      if(cruse_memory == 0) cruse_memory = 25;
      if(cruse_memory > 0) cruse_output = SPIN_RIGHT ;
      if(cruse_memory > 6) cruse_output = FWD;
          
   }
   
 //-----------------------------------------------------------------------
   #define AMIN 6
   #define HIRATIO 100
   #define LORATIO 40
   void chase()
   {  char max,dir,range;
      
      max = 0; dir = 0; range = 255;
      chase_output = 0;
      set_bit(PORTC,LED);   //Attack indictor LED off
      clear_bit(flags,attack);

      if((lf_avg < AMIN) 
         && (rf_avg < AMIN) 
         && (rr_avg < AMIN) 
         && (lr_avg < AMIN)) return;
     
      if(lf_range < range) { range = lf_range; dir = 1;}    //Find closest target
      if(rf_range < range) { range = rf_range; dir = 2;}
      if(lr_range < range) { range = lr_range; dir = 3;}
      if(rr_range < range) { range = rr_range; dir = 4;}

      
      
      switch(dir){
      case 0:  chase_output = 0;
               break;

      case 1:  
               //Target is in front
      case 2:  
               chase_output = FWD;
               if (lf_avg > rf_avg) {
                  
                  
                  if(f_ratio > LORATIO) chase_output = FWD_LEFT;
                  if(f_ratio > HIRATIO) chase_output = FWD_HARD_LEFT;
                  
               }
               else { 
                     if(f_ratio > LORATIO) chase_output = FWD_RIGHT;
                     if(f_ratio > HIRATIO) chase_output = FWD_HARD_RIGHT;
                     
               }

               
               if(range == 0){ 
                  chase_output = FWD_MAX;   //Attack mode
                  clear_bit(PORTC,LED);
                  set_bit(flags,attack);
               }else set_bit(PORTC,LED);
               
               break;


      case 3:
      case 4:  //Target is behind
               chase_output = REV;
               if(lr_avg > rr_avg) {
                  
                  if(r_ratio > LORATIO) chase_output = REV_LEFT;
                  if(r_ratio > HIRATIO) chase_output = REV_HARD_LEFT;
                  
               }
               else { 
                     if(r_ratio > LORATIO) chase_output = REV_RIGHT;
                     if(r_ratio > HIRATIO) chase_output = REV_HARD_RIGHT;
                     
               }

               if(range == 0){
                   chase_output = REV_MAX;    //Attack mode
                   clear_bit(PORTC,LED);
                   set_bit(flags,attack);
               }else set_bit(PORTC,LED);

               break;


      }

      
   }

   //-----------------------------------------------------------------------
   char escape_maneuver()
   {  char rc;

      rc = 0;
      if(flags & (1 << direction)){   //Going forward when tilted ?
         if(runTimer_L & 1)
              rc = REV_LEFT;  //Random left or right evasive maneuver
         else rc = REV_RIGHT;

      }else{                          //else going backwards...
         if(runTimer_L & 1)           //so switch to forward
           rc = FWD_LEFT;    //Random left or right evasive maneuver
         else rc = FWD_RIGHT;

      }
      return rc;
   }

   //-----------------------------------------------------------------------
  
   
   /* Tilt switch handler.  If tilt switch changes state
   from normal to tilt then reverse the robots direction.
   Do nothing if switch remains tilted or returns to normal.
   ONLY the transition from normal to tilted causes direction
   change.
   */
  void tilt()
  {
     static char last_tilt;
     char this_tilt, action;

     

    action = 0;
    this_tilt = tilt_dir & (1 << TILT);   //low true when bot is tilted

    if((this_tilt == 0) && (last_tilt == (1 << TILT))){
        
        action = escape_maneuver();
    }

    last_tilt = this_tilt;

    if(tilt_memory) return;

    if(action){
        tilt_output = action;
        tilt_memory = 20;
    }

    if(tilt_memory == 0) tilt_output = 0;


  }

  //------------------------------------------------------------------------
  //check to see if we are being pushed backward
  void pushed_back()
  {  char rc = 0;
     char action;
     action = 0;

     

     if(flags & ( 1 << attack)){   //In attack mode ?
        
        if(flags & (1 << direction)){   //Supposed to be going forward
           if (tilt_dir & (1 << DIR))   //Are we being pushed backward?
               action = escape_maneuver();   //yes, attempt escape
               
           
        }else{                          //Supposed to be going backward
               if(~tilt_dir & (1 << DIR)) //Are being pushed forward?
                   action = escape_maneuver();   //yes, attempt escape.
                  
        }

        
     }

    if(pushback_memory) return;

    if(action){
        pushed_back_output = action;
        pushback_memory = 20;         //Let this action persist for 2 secs.
    }

    if(pushback_memory == 0) pushed_back_output = 0;


     
  }    

      
 
   //------------------------------------------------------------------------
   //Stop motors while current exceeds I .
   void limit_motor_current(char I){

       if(drive_current > I){   
          current_limit_output = STOP;
       }else{
          current_limit_output = 0;
       }

   }

   //-------------------------------------------------------------------------
   //Stop motors while battery reads less than 5.5 volts (or whatever "low_limit" is)
   void check_battery(char low_limit){
      if((battery < low_limit) && (runTimer_L > 6)){ //After 6 secs and bat less than 5.5 volts.
         check_battery_output = STOP;
      } else{
         check_battery_output = 0;
      }

   }

   //----------------------------------------------------------------------
   char process_sonar()
   {  char rc=0;
      if(spi_stat){
         lf_avg = lf_ampl;
         rf_avg = rf_ampl;
         lr_avg = lr_ampl;
         rr_avg = rr_ampl;
         spi_stat = 0;
         f_ratio = ratio(lf_avg,rf_avg);
         r_ratio = ratio(lr_avg,rr_avg);
         rc = 255;
      }
      return rc;
   }

   //------------------------------------------------------------------------

  // Subsumption arbitration happens here.
   void arbitrate(){

      
     arb_code = 0;
     if(cruse_output) { arb_code = 1; dir_ctl = cruse_output; }                 //Lowest priority
     if(chase_output) { arb_code = 2; dir_ctl = chase_output; }
     
     if(pushed_back_output)  { 
        arb_code = 4; 
        dir_ctl = pushed_back_output;
        
     }

     if(tilt_output)  { arb_code = 3; dir_ctl = tilt_output;  }

     if(border_output) { 
        arb_code = 5; 
        dir_ctl = border_output;
        pushback_memory = 0;
        tilt_memory = 0;

     }

     if(current_limit_output) { arb_code = 6; dir_ctl = current_limit_output;}
     if(check_battery_output) { arb_code = 7; dir_ctl = check_battery_output; } //Highest priority
     
     dir_control(dir_ctl);   //Now tell the motors what to do.
     
   }

   //-------------------------------------------------------------------------
     
   
   main()
   {
      char i;
      char x;
      char y;
      char z;
      
    
      test_hi = 0;

      INTCON = 0;
      STATUS = 0;
      PORTA  = 0;
      PORTB  = 0;
      PORTC  = 0;
      T1CON  = 0;
      TMR0   = 0;
      workB  = 0x20;
      

      tx_ptr = 0;
      rx_ptr = 0;
      spi_ptr = 0;
      
      RCSTA = 0x90;             //SPEN + CREN (configure serial port)


      set_bit(STATUS,RP0);     //Bank1
      

      set_tris_b(0x00);        //Port B bits are outputs
      set_tris_c(0xd9);        //Port C outputs except 6:7 are UART pins and 3:4 are SPI pins, bit 0 is tilt sw inp
      TRISA = 0xff;            //port A all analog inputs
      
      
      OPTION_REG = 0x01;       //Prescaler 1:4 assigned to Timer 0, pullups on portB
      ADCON1 = 0;              //Set all port A to analog with VREF=VCC
      TXSTA = 0xa2;            //Set tx usart status/ctrl reg= CSRC + TXEN + TRMT
      SPBRG = BAUDDIV;         //Set baudrate
      SSPSTAT = 0x00;          //SPI PORT   slave mode CKE = 0  SMP = 0
      PIE1 = 0;                //Disable all peripheral interrupts
      set_bit(PIE1,SSPIE);     //Enable SPI port interrupts
      clear_bit(STATUS,RP0);   //Bank0
      
      SSPCON = 0;
      SSPCON = 0x15;            // SPI slave mode -SS disabled  clock = SCK pin  CKP=1
      set_bit(SSPCON,SSPEN);    //Enable SPI port

      clear_bit(INTCON,T0IF);  //Clear Timer 0 interrupt flag
      TMR0   = TMR0DIV;        //Start Timer 0
      set_bit(INTCON,T0IE);    //Enable Timer 0 interrupts
      set_bit(INTCON,PEIE);   //Enable peripheral interrupts
      set_bit(INTCON,GIE);     //Enable global interrupts
      iavg[0] = 0;
      iavg[1] = 0;
      iavg[2] = 0;
      iavg[3] = 0;
      drive_current_avg = 0;
      txstate = 0;

      

      clear_bit(PORTC,LED);     //indicator LED on
      //clear_bit(PORTC,IR_ENABLE);     //IR sensor LEDs on for testing
      testmode = 0;
      startup_timer = 50;
      tilt_memory = 0;
      border_memory = 0;
      pushback_memory = 0;
      cruse_memory = 0;
      timer2 = 0;
      flags = 0;
      divide10 = 0;
      divide100 = 0;
      counter1 = 10;
      runTimer_H = 0;
      runTimer_L = 0;
      left_speed = 0;
      right_speed = 0;
      r_integrate = 0;
      l_integrate = 0;
      r_speed = 0;
      l_speed = 0;
      speed = 0;
      lf_rr_sens = 127;
      rf_lr_sens = 127;
      
      bot_mode = MODE_STOP;
      dir_ctl = STOP;
      mon_dir_ctrl = dir_ctl;
      arb_code = 0;

      

      
      while(process_sonar() == 0);     //Wait for a packet from the sonar module
      
      
      

      /* Wait here for 5 seconds after power up */
      startup_timer = 50;             //5 sec
      while(startup_timer > 0){               //Wait 5 sec before starting
         if(startup_timer & 4) clear_bit(PORTC,LED);
            else set_bit(PORTC,LED); // blink the LED during the 5 sec wait.
         get_AD_values();
         if (txstate == 0) make_info_pkt(); //Get new data for transmission
         send_tx_buffer();  //Send data on RF link if TX ready

         if(~tilt_dir & (1 << TILT)) testmode = 1;  //If bot is started while lifed off ground don't run motors 
            else testmode = 0;
         
      }

      set_bit(PORTC,LED);  //Indicator LED off
      bot_mode = MODE_RUN; //Allow motors to run
      
      /* Start of the main program loop */

      for(;;){
      

       if (txstate == 0) make_info_pkt(); //Get new data for transmission
       send_tx_buffer();  //Send data on RF link if TX ready

       

       if(timer2 == 0){
           get_AD_values(); //Every 10ms read the A to D converters
           timer2 = 1;
       }

       /* Now check all the sensors */
       
       check_border();
       cruse();
       chase();
       tilt();
       pushed_back();
       limit_motor_current(20);  // 400 MA Limit
       check_battery(93);        //5.5 volt lower limit

       // Sort it all out and decide what to tell the motor controller.
       arbitrate();
       
        

               adjust_motor_speed();
        
        process_sonar();
        
        if(flags & (1 << tick100)){    //Every 100 ms do this...
             iavg[3] = iavg[2];        //Average the motor current over 400ms
             iavg[2] = iavg[1];
             iavg[1] = iavg[0];
             iavg[0] = drive_current;
             drive_current_avg = 0;
             for(i=0;i<4;i++) drive_current_avg = drive_current_avg + iavg[i];
             drive_current_avg = drive_current_avg >> 2;  //div by 4
             clear_bit(flags,tick100);
        } 
      
      }   //end main loop

   


} //End main()
/* End of source */

